/*
 * tmos - config.c
 *
 * POMMAX2 configuration
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/port.h>
#include "pommax2.h"

#define POMMAX2_CONFIG_SIZE	256
#define POMMAX2_COPIES		4

static __highmem u16 regs[POMMAX2_REG_MAX + 1];

u16 pommax2_config_read(unsigned char reg) __reentrant
{
	if (reg > POMMAX2_REG_MAX)
		return 0;
	return regs[reg];
}

void pommax2_config_write(u16 v, unsigned char reg) __reentrant
{
	if (reg <= POMMAX2_REG_MAX)
		regs[reg] = v;
}

void pommax2_default_config(void) __reentrant
{
	unsigned char i;

	for (i = 0; i <= POMMAX2_REG_MAX; i++)
		regs[i] = 0;

	regs[0] = POMMAX2_CONFIG_MAGIC | POMMAX2_CONFIG_VERSION
		| POMMAX2_REG_MAX;
}

#ifdef CONFIG_POMMAX2_NVCONFIG
static u16 pommax2_config_checksum(void) __reentrant
{
	u16 csum = 0;
	unsigned char i;

	for (i = 0; i <= POMMAX2_REG_MAX; i++)
		csum += regs[i];
	return csum;
}

void pommax2_config_load(void) __reentrant
{
	port_t fd;
	unsigned char c;
	unsigned char i;
	unsigned char nregs;

	fd = port_open(CONFIG_POMMAX2_NVCONFIG_PATH, O_WRONLY);
	if (fd == (port_t) -1)
		return;

	for (c = 0; c < POMMAX2_COPIES; c++) {
		u16 off = c * POMMAX2_CONFIG_SIZE;
		u16 v;

		pommax2_default_config();

		port_pread(fd, (void __kernel *) &v, 2, off);
		if ((v & POMMAX2_CONFIG_MAGIC_MASK) != POMMAX2_CONFIG_MAGIC)
			continue;
		if ((v & POMMAX2_CONFIG_VERSION_MASK) != POMMAX2_CONFIG_VERSION)
			continue;

		/*
		 * if the configuration stored in nvram is newer, ignore it.
		 */
		nregs = v & POMMAX2_CONFIG_REGS_MASK;
		if (nregs > POMMAX2_REG_MAX)
			continue;

		regs[0] = v;
		for (i = 1; i <= nregs; i++) {
			port_pread(fd, (void __kernel *) &v, 2, off + 2 * i);
			regs[i] = v;
		}

		port_pread(fd, (void __kernel *) &v, 2, off + 2 * i);

		if (v == pommax2_config_checksum())
			return;
	}
	pommax2_default_config();
}

char pommax2_config_save(void) __reentrant
{
	port_t fd;
	unsigned char c;
	unsigned char i;

	fd = port_open(CONFIG_POMMAX2_NVCONFIG_PATH, O_WRONLY);
	if (fd == (port_t) -1)
		return -1;

	regs[0] = POMMAX2_CONFIG_MAGIC | POMMAX2_CONFIG_VERSION
		| POMMAX2_REG_MAX;

	for (c = 0; c < POMMAX2_COPIES; c++) {
		u16 off = c * POMMAX2_CONFIG_SIZE;
		u16 v;

		for (i = 0; i <= POMMAX2_REG_MAX; i++) {
			v = regs[i];
			port_pwrite(fd, (void __kernel *) &v, 2, off + 2 * i);
		}

		v = pommax2_config_checksum();
		port_pwrite(fd, (void __kernel *) &v, 2, off + 2 * i);
	}
	return 0;
}
#endif
