/*
 * tmos - gain.c
 *
 * Analog scaling
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/jiffies.h>
#include "pommax2.h"

/*
 * POMMAX2 software analog scaling
 *
 * Some systems that use POMMAX2 cards needs some specific precise quantum.
 * Because the quantum on ADC is constant this can be performed only
 * by changing gain of the analog part of POMMAX2 card. The input gain
 * of each channel is configured by a set of 5 resistors so any gain change
 * usually requires replacing a large number of registers. Additionally
 * for each specific needed gain the proper resistances must be calculated.
 *
 * This module implements software gain to allow some adjustments of
 * gain. The gain is configurable in [-8; 8) range with 15-bit resolution.
 * For gains in range (-0.5; 0.5) the resolution is not sufficient
 * to archive 0.1% relative gain setting error.
 */

/*
 * 16-bit gain values configured by user
 */
static s16 __highmem gain[CHANNELS];

/*
 * 16-bit offset fixes configured by the user
 */
static s16 __highmem ofix[CHANNELS];

#if CONFIG_I51 && CONFIG_OPTIMIZE_SPEED
static u16 umul_shift12(u16 a, u16 b)
{
	u16 y;
	u16 t;
	u8 xh;
	u8 xl;
	u8 wh;
	u8 wl;

	xh = a >> 8;
	xl = a;
	wh = b >> 8;
	wl = b;

	y = xh * wh;
	if (y & 0xfc00)
		return 0x3fff;

	t = xh * wl;
	t += xl * wh;
	t += xl * wl >> 8;

	y <<= 4;
	y += t >> 4;

	if (y & 0xc000)
		return 0x3fff;
	return y;
}

static int16_t mul_shift12(int16_t a, int16_t b) __reentrant
{
	if (a < 0) {
		if (b < 0)
			return umul_shift12(-a, -b);
		else
			return -umul_shift12(-a, b);
	}
	if (b < 0)
		return -umul_shift12(a, -b);
	return umul_shift12(a, b);
}
#else
static int16_t mul_shift12(int16_t a, int16_t b)
{
	long y;

	y = (long) a * b >> 12;
	if (y > 0x3fff)
		y = 0x3fff;
	else if (y < -0x4000)
		y = -0x4000;
	return y;
}
#endif

int16_t pommax2_gain_apply(int16_t v, unsigned char ch) __reentrant
{
	s16 o = ofix[ch];

	/*
	 * add ofix with saturation. Quite ugly implementation, but
	 * it generates much better code on i51 with sdcc.
	 */
	if (v >= 0) {
	       	if (o >= 0) {
			v += o;
			if (v < 0)
				v = INT_MAX;
			goto out;
		}
	} else {
		if (o < 0) {
			v += o;
			if (v >= 0)
				v = INT_MIN;
			goto out;
		}
	}
	v += o;
out:
	return mul_shift12(v, gain[ch]);
}

void pommax2_gain_configure(void) __reentrant
{
	unsigned char i;
	s16 x;

	for (i = 0; i < CHANNELS; i++) {
		x = pommax2_config_read(POMMAX2_REG_GAIN_0 + i);

		/*
		 * gain uses full 16-bit range, but the config interface
		 * uses only 15-bit registers
		 */
		gain[i] = (x << 1) >> 1;

		x = pommax2_config_read(POMMAX2_REG_OFIX_0 + i);
		ofix[i] = x << 1;
	}
}
