/*
 * tmos - context.h
 *
 * Copyright (C) 2009  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __ASM_TASK_H
#define __ASM_TASK_H

#include <tmos/kernel.h>
#include <setjmp.h>
#include <tmos/jiffies.h>

typedef char pid_t;

#if CONFIG_MULTITASKING

#define TASK_RUNNING		0
#define TASK_INTERRUPTIBLE	1
#define TASK_UNINTERRUPTIBLE	2
#define TASK_DEAD		3
#define TASK_SLEEP		64

#define MAX_SCHEDULE_TIMEOUT	(256 - TASK_SLEEP)

#define set_task_state(task, state_value) \
	do { (task)->state = (state_value); } while (0)
#define set_current_state(state_value) \
	set_task_state(current, (state_value))
#define task_is_running(task) \
	((task)->state == TASK_RUNNING)
#define task_is_dead(task) \
	((task)->state == TASK_DEAD)

/*
 * thread context - each thread have this context structure.
 */
struct context {
	/*
	 * pointer to context that should preempt this context, for
	 * simple round-robin scheduling contexts can be linked
	 * circular list.
	 */
	struct context __kernel *next;
	unsigned char sp;		/* stack pointer */
#ifdef CONFIG_TASK_TIMES
	jiffies_t cpu_time;		/* used cpu time */
#endif

	unsigned char state;
};
#define TDATA_OFFSET	1

extern struct context task[CONFIG_TASKS_MAX];

unsigned char __unsupported_divisor(void);

/*
 * The division by 1, 2, 4, 8 and 16 is trivial. It's accurate in full
 * 8-bit range.
 *
 * The division by 3 and 5 is accurate up to 15 tasks.
 */
#define u8_div_const(a, b) \
	(((b) == 1) ? (a) : \
	 ((b) == 2) ? (a) >> 1 : \
	 ((b) == 3) ? (unsigned char) (5 * ((a) + 3)) >> 4 : \
	 ((b) == 4) ? (a) >> 2 : \
	 ((b) == 5) ? (unsigned char) (3 * ((a) + 5)) >> 4 : \
	 ((b) == 8) ? (a) >> 3 : \
	 ((b) == 16) ? (a) >> 4 : \
	 __unsupported_divisor())

#define task_tdata(tsk) \
	(u8_div_const((unsigned char) tsk - (unsigned char) task, \
			 sizeof(*tsk)) + TDATA_OFFSET)

#else
struct context {
	char dummy;
};
#endif

#if CONFIG_MULTITASKING
/* FIXME: this is probably broken */
#define GET_WCHAN(x)	(be16_to_cpu(*((uint16_t __data *) ((x)->sp - 4))))
#else
#define GET_WCHAN(x)	(0)
#endif

#define stack_alloc(size) ((unsigned char) iram_alloc((size)))

#if CONFIG_MULTITASKING
extern __bit __timers_pending;

#define timers_pending()	(__timers_pending)
#define set_timers_pending()	\
	do { \
		__timers_pending = 1; \
		raise_softirq(); \
	} while (0)
#define clear_timers_pending()	do { __timers_pending = 0; } while (0)
#else
#define timers_pending()	(0)
#define set_timers_pending()	do { } while (0)
#define clear_timers_pending()	do { } while (0)
#endif

#if CONFIG_USE_IDLE || CONFIG_MULTITASKING
extern __bit __need_resched;

#define need_resched()		(__need_resched)
#define set_need_resched()	do { __need_resched = 1; } while (0)
#define clear_need_resched()	do { __need_resched = 0; } while (0)
void idle(void);

#else
#define need_resched()		(0)
#define set_need_resched()	do { } while (0)
#define clear_need_resched()	do { } while (0)
#define idle()			do { } while (0)
#endif

#ifdef CONFIG_TDATA
#define TASK_SIZE	128
#define task_stack(pid)		(TDATA_BASE)
#else
#define TASK_SIZE	64
#define task_stack(pid)	(256 - pid * TASK_SIZE)
#endif

#if CONFIG_MULTITASKING

#if CONFIG_TDATA

#define WANT_PUSH_DP		1

#if CONFIG_ADUC841
#include <asm/aduc841.h>

#define switch_mm_copy_byte() 						\
	do {								\
		__asm							\
			movx	A, @DPTR				\
		__endasm;						\
		__asm							\
			xch	A, @r0					\
		__endasm;						\
		__asm							\
			movx	@DPTR, A				\
		__endasm;						\
		__asm							\
			inc	r0					\
		__endasm;						\
	} while (0)

#define switch_mm()							\
	do {								\
		register unsigned char old_page;			\
		register unsigned char new_page;			\
									\
		/* switch current and remember data pages */		\
		old_page = task_tdata(current);				\
		current = next_context;					\
		new_page = task_tdata(current);				\
									\
		DPH = old_page;						\
		DPL = 0;						\
									\
		DPCON = DPCON_AUTO | DPCON_SP_INC | DPCON_MP_INC	\
				| DPCON_DPSEL;				\
		DPH = new_page;						\
									\
		DPL = 0x80;						\
		__asm							\
			mov	r0, DPL					\
		__endasm;						\
		DPL = 0;						\
		__asm							\
01000$:									\
		__endasm;						\
		switch_mm_copy_byte();					\
		switch_mm_copy_byte();					\
		switch_mm_copy_byte();					\
		switch_mm_copy_byte();					\
									\
		switch_mm_copy_byte();					\
		switch_mm_copy_byte();					\
		switch_mm_copy_byte();					\
		switch_mm_copy_byte();					\
									\
		__asm							\
			mov	a, r0					\
		__endasm;						\
		__asm							\
			jnz	01000$					\
		__endasm;						\
		DPCON = 0x00;						\
	} while (0)
#endif

#define STORE_MEM() \
	do { \
		register unsigned long __data *s \
			= (unsigned long __data *) 0x80; \
		register unsigned long __xdata *d \
			= (unsigned long __xdata *) \
			(task_tdata(current) << 8); \
		do { \
			*d++ = *s++; \
			*d++ = *s++; \
		} while (s); \
	} while (0)

#define LOAD_MEM() \
	do { \
		register unsigned long __data *d \
			= (unsigned long __data *) 0x80; \
		register unsigned long __xdata *s \
			= (unsigned long __xdata *) \
			(task_tdata(current) << 8); \
		do { \
			*d++ = *s++; \
			*d++ = *s++; \
		} while (d); \
	} while (0)

#else

#if CONFIG_PREEMPT
#define WANT_PUSH_DP		1
#endif

#define STORE_MEM()	do { } while (0)
#define LOAD_MEM()	do { } while (0)
#endif

#if WANT_PUSH_DP
#define PUSH_DP() \
	do { \
		__asm__(push dpl;); \
		__asm__(push dph;); \
	} while (0)
#define POP_DP() \
	do { \
		__asm__(pop dph;); \
		__asm__(pop dpl;); \
	} while (0)
#else
#define PUSH_DP()		do { } while (0)
#define POP_DP()		do { } while (0)
#endif

#define PUSH_REG(x) \
	do { \
		__asm__(__TEMPREG = x;); \
		__asm__(push __TEMPREG;); \
	} while (0)

#define POP_REG(x) \
	do { \
		__asm__(__TEMPREG = x;); \
		__asm__(pop __TEMPREG;); \
	} while (0)

#if CONFIG_PREEMPT
#define PUSH_REGS() \
	do { \
		PUSH_REG(0x00); \
		PUSH_REG(0x01); \
		PUSH_REG(0x02); \
		PUSH_REG(0x03); \
		PUSH_REG(0x04); \
		PUSH_REG(0x05); \
		PUSH_REG(0x06); \
		PUSH_REG(0x07); \
	} while (0)

#define POP_REGS() \
	do { \
		POP_REG(0x07); \
		POP_REG(0x06); \
		POP_REG(0x05); \
		POP_REG(0x04); \
		POP_REG(0x03); \
		POP_REG(0x02); \
		POP_REG(0x01); \
		POP_REG(0x00); \
	} while (0)
#else
#define PUSH_REGS()	do { } while (0)
#define POP_REGS()	do { } while (0)
#endif

#define irq_preempt_enabled() preempt_enabled()

#ifndef switch_mm
#define switch_mm()							\
	do {								\
		STORE_MEM();						\
		current = next_context;					\
		LOAD_MEM();						\
	} while (0)
#endif

/*
 * TASK_SWITCH - do context switch
 *
 * This macro should be used in interrupt, thread_switch() should
 * be used in other contexts. In most cases this macro should be called
 * via irq_switch_task().
 *
 * Make sure than stack is correct, before TASK_SWITCH() is should have:
 *
 * 	RETURN_ADDR_0
 * 	RETURN_ADDR_1
 * 	PSW
 * 	[DPL]	- only in TDATA model on !ADUC841
 * 	[DPH]	- only in TDATA model on !ADUC841
 * 	ACC
 *
 * Also preemption should be possible:
 * 	- preemption should be enabled
 * 	- current task should be in preemptive section
 */
#define TASK_SWITCH() \
	do { \
		PUSH_DP(); \
		PUSH_REGS(); \
		current->sp = SP; \
		switch_mm(); \
		SP = current->sp; \
		POP_REGS(); \
		POP_DP(); \
	} while (0)

extern struct context __kernel *next_context;

/**
 * __schedule - the real scheduler
 *
 * This function is called in softirq context and it's responsible to
 * select next context to run. The result is stored in next_context
 * global variable.
 */
#if CONFIG_SCHED_RR
#define __schedule() \
	do { \
		next_context = current->next; \
	} while (0)
#elif CONFIG_SCHED_RT_RR
#define __schedule() \
	do { \
		if (task_is_running(&task[0])) \
			next_context = &task[0]; \
		else \
			next_context = current->next; \
	} while (0)
#else
void __schedule(void) __using(SOFTIRQ_BANK);
#endif

void do_timers(void) __using(SOFTIRQ_BANK);

/**
 * switch_task - switch task
 *
 * See TASK_SWITCH().
 */
#define switch_task() \
	do { \
		clear_need_resched(); \
		__schedule(); \
		if (next_context != current) \
			TASK_SWITCH(); \
	} while (0)

#ifdef CONFIG_PREEMPT
/**
 * irq_switch_task - switch task from interrupt.
 *
 * This function should be called after waking up process that should
 * preempt current process. It's also called in periodic timer interrupt
 * to implement round-robin scheduling.
 *
 * See TASK_SWITCH().
 */
#define irq_switch_task() \
	do { \
		if (irq_preempt_enabled() && need_resched()) \
			switch_task(); \
	} while (0)
#else
#define irq_switch_task()	do { } while (0)
#endif

/**
 * current - current context
 */
extern struct context __kernel *current;

char context_init(struct context __kernel *context) __preempt;
char thread_create(struct context __kernel *context, unsigned char sp)
	__need_lock(PREEMPT) __reentrant;
void thread_switch(void);
void show_context_state(struct context __kernel *context);

#else
#define irq_switch_task()	do { } while (0)
#endif

#endif
